import java.io.*;
import java.util.*;
import java.util.function.Function;

public class TestGen {
    int testCount, subtaskCount;
    PrintWriter testsDesc;
    Random random = new Random(0x1_1_1_C_A);
    public static final int MAX_N = 100_000;//10^5
    public static final int MAX_VAL = 1_000_000_000 - 1;
    public static final int MAX_LEN = 9;

    private String folder;
    private String subtask;
    private int groupNo;

    public void setFolder(String folder) {
        this.folder = folder;
    }

    public void setSubtask(String subtask) {
        this.subtask = subtask;
    }

    public void genSamples() throws IOException {
        setFolder("preliminary");
        setSubtask("subtask0");

        System.out.println("==== Generating samples ===");
        print(new Test(new int[]{10, 100, 23, 1, 60}, new String[]{
                "?",
                "0?1",
                "70?",
                "6?",
                "00"
        }));

        print(new Test(new int[]{1, 2, 3, 66, 10, 20, 34, 48, 50, 103}, new String[]{
                "2?1",
                "0?1",
                "???",
                "??0",
        }));

        testCount = 0;
    }

    public void genHandleTests(int mx, int points) throws IOException {

    }


    class Generator {
        final int SZ = 200;
        int[] a = new int[SZ];
        int ptr = 0;
        int[] suffix = new int[SZ];
        int[] suffPow = new int[SZ];
        private String description;

        Generator(Function<Integer, Integer[]> arrayGenerator, String description) {
            this.description = description;
            Integer[] arr = arrayGenerator.apply(SZ);
            for (int i = 0; i < SZ; ++i) {
                a[i] = arr[i];
                suffix[i] = 1;
                suffPow[i] = 1;
            }
        }

        public int nextQuery() {
            return a[ptr++];
        }

        public int nextAi() {
            int id = random.nextInt(ptr);
            if (suffix[id] == suffPow[id]) {
                suffPow[id] *= 10;
                suffix[id] = 0;
            }
            int ret = (a[id] / suffPow[id]) * suffPow[id] + suffix[id];
            suffix[id]++;
            return ret;
        }

        public String description() {
            return description;
        }
    }

    Generator randomGenerator = new Generator(n-> {
        Integer[] ret = new Integer[n];
        for (int i = 0; i < n; ++i)
            ret[i] = randInterval(MAX_VAL / 10 + 1, MAX_VAL);
        return ret;
    }, "RandomGenerator228");

    Generator randomGeneratorRepeat = new Generator(n-> {
        Integer[] ret = new Integer[n];
        for (int i = 0; i < n; ++i)
            ret[i] = randInterval(MAX_VAL / 10 + 1, MAX_VAL);
        return ret;
    }, "RandomGeneratorRepeat") {
        @Override
        public int nextAi() {
            int id = random.nextInt(ptr);
            return a[id];
        }
    };

    Generator sequenceGenerator = new Generator(n->{
        int start = randInterval(MAX_VAL / 10 + 1, MAX_VAL / 2);
        Integer[] ret = new Integer[n];
        for (int i = 0; i < n; ++i)
            ret[i] = start + i;
        return ret;
    }, "SequenceGenerator239");

    public void genSubtask(int n, int m, int maxVal, int len, int quest, int groupNo, int points) throws IOException {
        ++subtaskCount;
        String subtask = "subtask" + subtaskCount;
        setSubtask(subtask);
        setFolder("tests");
        setGroupNo(groupNo);
        System.out.println("==== Generating group #" + groupNo + " ===");


        print(onlyDigits(n, m, maxVal, len));

        if (m != 1)
            withLeadingZeros(n, maxVal, len);

        if (quest != 1) {
            print(allQuest(n, m, maxVal / 10 + 1, maxVal, len, len));
            print(allQuest(n, m, maxVal / 10 + 1, maxVal, 1, len));
            print(allQuest(n, m, 1, 1, 1, len));

            print(allQuestSmallMix(n, m, maxVal / 10 + 1, maxVal, len, len));
            print(allQuestSmallMix(n, m, 1, maxVal, 1, len));
            print(allQuestSmallMix(n, m, 1, 1, 1, len));
        }

        double[] prob = new double[]{0, 0.3, 0.5, 0.8, 1.0};
        for (int i = 0; i < prob.length; ++i)
            if (quest == 1)
                print(genRandom1(n, m, maxVal, len, prob[i]));
            else
                print(genRandom2(n, m, maxVal, len, prob[i]));
        if (quest != 1) {
            print(genRandom2(n, m, maxVal, len, 1.0 / 11));
            print(genRandom2(n, m, maxVal, len, 1.0 / 11));
        }

        if (groupNo != 3) {
            print(genSameNumbers(n, m, 0, maxVal, quest, false));
            print(genSameNumbers(n, m, 0, maxVal, quest, true));
            print(genSameNumbers(n, m, 0.2, maxVal, quest, false));
            print(genSameNumbers(n, m, 0.2, maxVal, quest, true));
            print(genSameNumbers(n, m, 0.5, maxVal, quest, false));
            print(genSameNumbers(n, m, 0.5, maxVal, quest, true));
        }

        if (groupNo == 4) {
            print(genAllCombinations(n, m, randomGeneratorRepeat));
            print(genAllCombinations(n, m, randomGenerator));
            print(genAllCombinations(n, m, sequenceGenerator));
        }


        testsDesc.println(String.format("comment_subtask%d=\"n &lt;= %d, m &lt;= %d, len &lt;= %d, quest &lt;= %d\"", subtaskCount, n, m, len, quest));
        testsDesc.println(String.format("bonus_subtask%d=%d", subtaskCount, points));
    }

    private Test genSameNumbers(int n, int m, double prob, int maxVal, int quest, boolean isRand) {
        int[] a = new int[n];
        int start = randInterval(maxVal / 10 + 1, maxVal);
        for (int i = 0; i < n; ++i)
            if (isRand)
                a[i] = randInterval(maxVal / 10 + 1, maxVal);
            else
                a[i] = Math.min(start + i, maxVal);
        shuffle(a);

        String[] q = new String[m];
        for (int i = 0; i < m; ++i) {
            q[i] = Integer.toString(a[random.nextInt(n)]);
            int cnt = 0;
            for (int j = 0; j < q[i].length(); ++j)
                if (cnt < quest && random.nextDouble() < prob) {
                    q[i] = q[i].substring(0, j) + "?" + q[i].substring(j + 1);
                    ++cnt;
                }
        }
        return new Test(a, q, String.format("same numbers n = %d, m = %d, prob = %f, isRand = %b", n, m, prob, isRand));
    }

    private Test genAllCombinations(int n, int m, Generator generator) {
        int sz = 0;
        String[] q = new String[m];
        int MASKS = 1<<MAX_LEN;
        while (sz < m) {
            int cur = generator.nextQuery();
            for (int i = 0; i < MASKS && sz < m; ++i) {
                String s = "";
                int x = cur;
                for (int j = 0; j < MAX_LEN; ++j) {
                    if (((i >> j) & 1) == 1)
                        s = "?" + s;
                    else
                        s = (char)(x % 10 + '0') + s;
                    x /= 10;
                }
                q[sz++] = s;
            }
        }
        int[] a = new int[n];
        for (int i = 0; i < n; ++i)
            a[i] = generator.nextAi();
        shuffle(q);
        return new Test(a, q, "brute by masks of numbers: " + generator.description());
    }

    private void shuffle(String[] q) {
        List<String> ls = Arrays.asList(q);
        shuffle(ls);
        for (int i = 0; i < q.length; ++i)
            q[i] = ls.get(i);
    }

    private void shuffle(int[] q) {
        Integer[] q2 = new Integer[q.length];
        for (int i = 0; i < q2.length; ++i)
            q2[i] = q[i];

        List<Integer> ls = Arrays.asList(q2);
        shuffle(ls);
        for (int i = 0; i < q.length; ++i)
            q[i] = ls.get(i);
    }

    private Test genRandom2(int n, int m, int maxVal, int len, double probably) {
        int[] a = new int[n];
        String[] q = new String[m];
        for (int i = 0; i < n; ++i)
            a[i] = randInt(maxVal);
        for (int i = 0; i < m; ++i) {
            int l = randInt(len);
            StringBuilder sb = new StringBuilder();
            for (int j = 0; j < l; ++j)
                if (random.nextDouble() < probably)
                    sb.append('?');
                else
                    sb.append(randDigit());
            q[i] = sb.toString();
        }
        return new Test(a, q, String.format("random test with many questions n = %d, m = %d, maxVal = %d, len = %d", n, m, maxVal, len));
    }

    private Test genRandom1(int n, int m, int maxVal, int len, double probably) {
        int[] a = new int[n];
        String[] q = new String[m];
        for (int i = 0; i < n; ++i)
            a[i] = randInt(maxVal);
        for (int i = 0; i < m; ++i) {
            int l = randInt(len);
            String s = randStringNumber(l);
            if (random.nextDouble() < probably) {
                int id = random.nextInt(l);
                s = s.substring(0, id) + "?" + s.substring(id + 1);
            }
            q[i] = s;
        }
        return new Test(a, q, String.format("random test with one question: n = %d, m = %d, maxVal = %d, len = %d", n, m, maxVal, len));
    }

    private String randStringNumber(int len) {
        String s = "";
        for (int i = 0; i < len; ++i)
            s += randDigit();
        return s;
    }

    private Test withLeadingZeros(int n, int maxVal, int len) {
        int[] a = new int[n];
        for (int i = 0; i < n; ++i)
            a[i] = randInt(maxVal);
        String[] q = new String[len];
        String s = "";
        for (int i = 0; i < len; ++i) {
            s += "0";
            q[i] = s;
        }
        return new Test(a, q, "with leading zeros");
    }

    private int pow10(int len) {
        int p = 1;
        for (int i = 1; i <= len; ++i)
            p *= 10;
        return p;
    }

    private Test onlyDigits(int n, int m, int maxValAi, int len) {
        int[] a = new int[n];
        String[] q = new String[m];
        for (int i = 0; i < n; ++i)
            a[i] = randInt(maxValAi);

        for (int i = 0; i < m; ++i) {
            int l = randInt(len);
            int x = randInt(pow10(l) - 1);
            String num = Integer.toString(x);
            int zeros = random.nextInt() % (len - l + 1);
            for (int p = 0; p < zeros; ++p)
                num = "0" + num;
            q[i] = num;
        }
        return new Test(a, q, "only digits");
    }

    private Test allQuest(int n, int m, int minVal, int maxVal, int minLen, int maxLen) {
        int[] a = new int[n];
        for (int i = 0; i < n; ++i)
            a[i] = randInterval(minVal, maxVal);
        String[] q = new String[m];
        String[] quest = new String[maxLen + 1];
        quest[0] = "";
        for (int i = 1; i <= maxLen; ++i)
            quest[i] = quest[i - 1] + "?";
        for (int i = 0; i < m; ++i)
            q[i] = quest[randInterval(minLen, maxLen)];
        return new Test(a, q, String.format("all questions n = %d, m = %d, minVal = %d, maxLen = %d, maxLen = %d, Ai in [%d, %d]", n, m, maxVal, minLen, maxLen, minVal, maxVal));
    }

    private Test allQuestSmallMix(int n, int m, int minVal, int maxVal, int minLen, int maxLen) {
        int[] a = new int[n];
        for (int i = 0; i < n; ++i)
            a[i] = randInterval(minVal, maxVal);
        String[] q = new String[m];

        for (int i = 0; i < m; ++i)
            q[i] = randChangeQuestions(randInterval(minLen, maxLen));
        return new Test(a, q, String.format("all questions small mix n = %d, m = %d, maxVal = %d, minLen = %d, maxLen = %d, Ai in [%d, %d]", n, m, maxVal, minLen, maxLen, minVal, maxVal));
    }

    private int randInterval(int minVal, int maxVal) {
        return random.nextInt(maxVal - minVal + 1) + minVal;
    }

    private String randChangeQuestions(int l) {
        int t = random.nextInt(Math.min(l, 3) + 1);
        int[] comb = genCombine(l, t);
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < l; ++i)
            sb.append('?');
        for (int i = 0; i < comb.length; ++i)
            sb.setCharAt(i, randDigit());
        return sb.toString();
    }

    private char randDigit() {
        return (char)(random.nextInt(10) + '0');
    }

    private int[] genCombine(int n, int k) {
        List<Integer> p = new ArrayList<>();
        for (int i = 0; i < n; ++i)
            p.add(i);
        shuffle(p);
        int[] ret = new int[k];
        List<Integer> sublist = p.subList(0, k);
        for (int i = 0; i < k; ++i)
            ret[i] = sublist.get(i);
        return ret;
    }

    private void shuffle(List<?> t) {
        Collections.shuffle(t, random);
    }

    private int randInt(int maxVal) {
        return random.nextInt(maxVal) + 1;
    }

    public void setGroupNo(int groupNo) {
        this.groupNo = groupNo;
    }

    public static class Test {
        public int n;
        public int m;
        public int[] a;
        public String[] q;
        public String descr;

        public Test(int[] a, String[] q) {
            this.a = a;
            this.q = q;
            n = a.length;
            m = q.length;
        }

        public Test(int[] a, String[] q, String descr) {
            this(a, q);
            this.descr = descr;
        }
    }

    public void print(Test cur) throws IOException {
        String path = String.format("../%s/%02d", folder, ++testCount);
        PrintWriter out = new PrintWriter(path);

        System.out.println("Generating test #" + testCount + ": " + cur.descr);

        out.println(String.format("%d %d", cur.n, cur.m));
        for (int i = 0; i < cur.n; ++i)
            if (i + 1 == cur.n)
                out.println(cur.a[i]);
            else
                out.print(cur.a[i] + " ");
        for (int i = 0; i < cur.m; ++i)
            out.println(cur.q[i]);

        out.close();
        testsDesc.println(String.format("%s, folder=%s, subtask=%s, points=%d, comment=\"%s\"", path, folder, subtask, 0, cur.descr));
    }

    public void run() throws IOException {
        new File("../preliminary").mkdir();
        new File("../tests").mkdir();
        testsDesc = new PrintWriter("../tests.desc");

        genSamples();
        genSubtask(1000, 1000, 1000 - 1, 3, 3, 1, 21);
        genSubtask(MAX_N, MAX_N, MAX_VAL, 9, 1, 2, 18);
        genSubtask(MAX_N, 1, MAX_VAL, 6, 6, 3, 23);
        genSubtask(MAX_N, MAX_N, MAX_VAL, MAX_LEN, MAX_LEN, 4, 38);

        testsDesc.close();
    }

    public static void main(String[] args) throws IOException {
        new TestGen().run();
    }
}
